"""
Solving problems with Dynamic Programming

Copyright 2020, University of Freiburg

Philipp Schneider <philipp.schneider@cs.uni-freiburg.de>
"""

import sys

sys.setrecursionlimit(10000)


def has_equal_partition(array):
    '''
        Returns True if and only if there is a partition of
        the integers in the input array that have equal sum
        of weights.

        Parameter:
            array - contains positive integers

        Returns:
            a boolean value representing whether this array
            can be partitioned into sets with equal sum
        >>> has_equal_partition([1,2,3,4])
        True
        >>> has_equal_partition([1,2,3,5])
        False
        >>> has_equal_partition(read_array("set1.txt"))
        True
        >>> has_equal_partition(read_array("set3.txt"))
        True
        >>> has_equal_partition(read_array("set2.txt"))
        False
    '''
    n = len(array)
    sum = 0
    for i in range(0,n):
        sum += array[i]
    
    if sum % 2 == 1 :
        return False
    
    memo = {}
    return has_equal_partition_rec(array,n-1,sum // 2, memo)

def has_equal_partition_rec(array, i,j, memo):
    if j == 0:
        return True
    if i == 0:
        return array[0] == j
    if j < 0 or i < 0:
        return False
    if (i, j) in memo:
        return memo[(i, j)]
    a = has_equal_partition_rec(array,i-1,j - array[i],memo) or has_equal_partition_rec(array,i-1,j,memo)
    memo[(i,j)] = a
    return a



def no_consecutive_ones(n):
    '''
        Returns the number of n-digit binary strings
        without consecutive ones

        Parameter:
            n - positive integer

        Returns:
            positive integer

        >>> [no_consecutive_ones(n) for n in [10,20,50]]
        [144, 17711, 32951280099]
    '''
    memo = {}
    return no_consecutive_ones_ending(n+1,0,memo)


def no_consecutive_ones_ending(n,i,memo):
    if (n,i) in memo:
        return memo[(n,i)]
    if n == 1 :
        a = 1
    elif i == 1:
        a = no_consecutive_ones_ending(n-1,0,memo)
    else:
        a = no_consecutive_ones_ending(n-1,0,memo) + no_consecutive_ones_ending(n-1,1,memo)
    memo[(n,i)] = a
    return a


def read_array(filename):
    '''
        reads an integer array from a file

        Parameter:
            filename - name of file

        Returns:
            array of integers in the file
    '''
    with open(filename) as input:
        strings = input.read().split(' ')
        array = [int(strings[i]) for i in range(len(strings))]
    return array
